@@ -34,20 +34,22 @@ static void assert_ssh(int rc, const char * what, ssh_session ssh){
   }
 }
 
-static size_t password_cb(SEXP rpass, const char * prompt, char buf[1024]){
+static size_t password_cb(SEXP rpass, const char * prompt, char buf[1024], const char *user){
   if(Rf_isString(rpass) && Rf_length(rpass)){
     strncpy(buf, CHAR(STRING_ELT(rpass, 0)), 1024);
     return Rf_length(STRING_ELT(rpass, 0));
   } else if(Rf_isFunction(rpass)){
     int err;
-    SEXP call = PROTECT(Rf_lcons(rpass, Rf_lcons(make_string(prompt), R_NilValue)));
+    SEXP question = PROTECT(make_string(prompt));
+    Rf_setAttrib(question, R_NamesSymbol, make_string(user));
+    SEXP call = PROTECT(Rf_lcons(rpass, Rf_lcons(question, R_NilValue)));
     SEXP res = PROTECT(R_tryEval(call, R_GlobalEnv, &err));
     if(err || !Rf_isString(res)){
-      UNPROTECT(2);
+      UNPROTECT(3);
       Rf_error("Password callback did not return a string value");
     }
     strncpy(buf, CHAR(STRING_ELT(res, 0)), 1024);
-    UNPROTECT(2);
+    UNPROTECT(3);
     return strlen(buf);
   }
   Rf_errorcall(R_NilValue, "unsupported password type");
@@ -55,19 +57,21 @@ static size_t password_cb(SEXP rpass, const char * prompt, char buf[1024]){
 
 int my_auth_callback(const char *prompt, char *buf, size_t len, int echo, int verify, void *userdata){
   SEXP rpass = (SEXP) userdata;
-  password_cb(rpass, prompt, buf);
+  password_cb(rpass, prompt, buf, "");
   return SSH_OK;
 }
 
-static int auth_password(ssh_session ssh, SEXP rpass){
+static int auth_password(ssh_session ssh, SEXP rpass, const char *user){
   char buf[1024];
-  password_cb(rpass, "Please enter your password", buf);
+  char prompt[1024];
+  snprintf(prompt, 1023, "Please enter ssh password for user '%s'", user ? user : "???");
+  password_cb(rpass, prompt, buf, user);
   int rc = ssh_userauth_password(ssh, NULL, buf);
   assert_ssh(rc == SSH_AUTH_ERROR, "password auth", ssh);
   return rc;
 }
 
-static int auth_interactive(ssh_session ssh, SEXP rpass){
+static int auth_interactive(ssh_session ssh, SEXP rpass, const char *user){
   int rc = ssh_userauth_kbdint(ssh, NULL, NULL);
   while (rc == SSH_AUTH_INFO) {
     const char * name = ssh_userauth_kbdint_getname(ssh);
@@ -80,7 +84,9 @@ static int auth_interactive(ssh_session ssh, SEXP rpass){
     for (int iprompt = 0; iprompt < nprompts; iprompt++) {
       char buf[1024];
       const char * prompt = ssh_userauth_kbdint_getprompt(ssh, iprompt, NULL);
-      password_cb(rpass, prompt, buf);
+      char question[1024];
+      snprintf(question, 1023, "Authenticating user '%s'. %s", user, prompt);
+      password_cb(rpass, question, buf, user);
       if (ssh_userauth_kbdint_setanswer(ssh, iprompt, buf) < 0)
         return SSH_AUTH_ERROR;
     }
@@ -90,19 +96,21 @@ static int auth_interactive(ssh_session ssh, SEXP rpass){
 }
 
 /* authenticate client */
-static void auth_any(ssh_session ssh, ssh_key privkey, SEXP rpass){
+static void auth_any(ssh_session ssh, ssh_key privkey, SEXP rpass, const char *user){
   if(ssh_userauth_none(ssh, NULL) == SSH_AUTH_SUCCESS)
     return;
   int method = ssh_userauth_list(ssh, NULL);
   if (method & SSH_AUTH_METHOD_PUBLICKEY){
     if(privkey != NULL && ssh_userauth_publickey(ssh, NULL, privkey) == SSH_AUTH_SUCCESS)
       return;
+    // ssh_userauth_publickey_auto() tries both ssh-agent and standard keys in ~/.ssh
+    // it also automatically picks up SSH_ASKPASS env var set by 'askpass' package
     if(privkey == NULL && ssh_userauth_publickey_auto(ssh, NULL, NULL) == SSH_AUTH_SUCCESS)
       return;
   }
-  if (method & SSH_AUTH_METHOD_INTERACTIVE && auth_interactive(ssh, rpass) == SSH_AUTH_SUCCESS)
+  if (method & SSH_AUTH_METHOD_INTERACTIVE && auth_interactive(ssh, rpass, user) == SSH_AUTH_SUCCESS)
     return;
-  if (method & SSH_AUTH_METHOD_PASSWORD && auth_password(ssh, rpass) == SSH_AUTH_SUCCESS)
+  if (method & SSH_AUTH_METHOD_PASSWORD && auth_password(ssh, rpass, user) == SSH_AUTH_SUCCESS)
     return;
   Rf_error("Authentication failed, permission denied");
 }
@@ -143,13 +151,33 @@ SEXP C_start_session(SEXP rhost, SEXP rport, SEXP ruser, SEXP keyfile, SEXP rpas
   size_t hlen = 0;
   assert_ssh(myssh_get_publickey(ssh, &key), "myssh_get_publickey", ssh);
   assert_ssh(ssh_get_publickey_hash(key, SSH_PUBLICKEY_HASH_SHA1, &hash, &hlen), "ssh_get_publickey_hash", ssh);
+#if LIBSSH_VERSION_MINOR < 8
   if(!ssh_is_server_known(ssh)){
-    Rprintf("New server fingerprint: %s\n", ssh_get_hexa(hash, hlen));
-    ssh_write_knownhost(ssh);
+    Rprintf("Server fingerprint: %s\n", ssh_get_hexa(hash, hlen));
   }
+#else
+  switch(ssh_session_is_known_server(ssh)){
+  case SSH_KNOWN_HOSTS_OK:
+    REprintf("Found known server key: %s\n", ssh_get_hexa(hash, hlen));
+    break;
+  case SSH_KNOWN_HOSTS_UNKNOWN:
+    REprintf("New server key: %s\n", ssh_get_hexa(hash, hlen));
+    ssh_session_update_known_hosts(ssh);
+    break;
+  case SSH_KNOWN_HOSTS_OTHER:
+  case SSH_KNOWN_HOSTS_CHANGED:
+    REprintf("Server key has changed (possible attack?!): %s\nPlease update your ~/.ssh/known_hosts file\n", ssh_get_hexa(hash, hlen));
+    break;
+  case SSH_KNOWN_HOSTS_ERROR:
+  case SSH_KNOWN_HOSTS_NOT_FOUND:
+    if(loglevel > 0)
+      REprintf("Could not load a known_hosts file\n");
+    break;
+  }
+#endif
 
   /* Authenticate client */
-  auth_any(ssh, privkey, rpass);
+  auth_any(ssh, privkey, rpass, user);
 
   /* display welcome message */
   char * banner = ssh_get_issue_banner(ssh);
